#ifndef __PLAT_TYPES_H__
#    define __PLAT_TYPES_H__

#    ifdef HAVE_CONFIG_H
#        include <config.h>
#    endif

#    include "sys/plat_posix.h"
#    include "sys/plat_windows.h"

#endif                          /* __PLAT_TYPES_H__ */
